"""Apply writegood tool and gather results.

Website: https://github.com/btford/write-good
"""

import logging
import re
import subprocess
from typing import List, Match, Optional, Pattern

from statick_tool.issue import Issue
from statick_tool.package import Package
from statick_tool.tool_plugin import ToolPlugin


class WriteGoodToolPlugin(ToolPlugin):  # type: ignore
    """Apply writegood tool and gather results."""

    def get_name(self) -> str:
        """Get name of tool."""
        return "writegood"

    def scan(self, package: Package, level: str) -> Optional[List[Issue]]:
        """Run tool and gather output."""
        tool_bin = "write-good"

        flags: List[str] = ["--parse"]
        user_flags = self.get_user_flags(level)
        flags += user_flags

        files: List[str] = []
        if "md_src" in package:
            files += package["md_src"]
        if "rst_src" in package:
            files += package["rst_src"]

        total_output: List[str] = []

        try:
            exe = [tool_bin] + flags + files
            output = subprocess.check_output(
                exe, stderr=subprocess.STDOUT, universal_newlines=True
            )
            total_output.append(output)

        except subprocess.CalledProcessError as ex:
            if ex.returncode in (
                0,
                255,
            ):  # writegood returns 0 or 255 upon linting errors
                total_output.append(ex.output)
            else:
                logging.warning("%s failed! Returncode = %d", tool_bin, ex.returncode)
                logging.warning("%s exception: %s", self.get_name(), ex.output)
                return None

        except OSError as ex:
            logging.warning("Couldn't find %s! (%s)", tool_bin, ex)
            return None

        for output in total_output:
            logging.debug("%s", output)

        with open(self.get_name() + ".log", "w", encoding="utf8") as fid:
            for output in total_output:
                fid.write(output)

        issues: List[Issue] = self.parse_output(total_output)
        return issues

    def parse_output(self, total_output: List[str]) -> List[Issue]:
        """Parse tool output and report issues."""
        writegood_re = r"(.+):(\d+):(\d+):(.+)"
        parse: Pattern[str] = re.compile(writegood_re)
        issues: List[Issue] = []

        for output in total_output:
            for line in output.splitlines():
                match: Optional[Match[str]] = parse.match(line)
                if match:
                    issues.append(
                        Issue(
                            match.group(1),
                            match.group(2),
                            self.get_name(),
                            "suggestion",
                            "1",
                            match.group(4),
                            None,
                        )
                    )

        return issues



"""Apply rst-lint tool and gather results."""

import logging
from typing import List, Optional

import restructuredtext_lint
from docutils.utils import SystemMessage

from statick_tool.issue import Issue
from statick_tool.package import Package
from statick_tool.tool_plugin import ToolPlugin


class RstlintToolPlugin(ToolPlugin):  # type: ignore
    """Apply rst-lint tool and gather results."""

    def get_name(self) -> str:
        """Get name of tool."""
        return "rstlint"

    # pylint: disable=too-many-locals
    def scan(self, package: Package, level: str) -> Optional[List[Issue]]:
        """Run tool and gather output."""
        flags: List[str] = []
        user_flags = self.get_user_flags(level)
        flags += user_flags

        files: List[str] = []
        if "rst_src" in package:
            files += package["rst_src"]

        total_output: List[SystemMessage] = []

        for src in files:
            output = restructuredtext_lint.lint_file(src, None, flags)
            total_output.extend(output)

        for output in total_output:
            logging.debug("%s", str(output))

        with open(self.get_name() + ".log", "w", encoding="utf8") as fid:
            for output in total_output:
                fid.write(str(output))

        issues: List[Issue] = self.parse_output(total_output)
        return issues

    # pylint: enable=too-many-locals

    def parse_output(self, total_output: List[SystemMessage]) -> List[Issue]:
        """Parse tool output and report issues."""
        issues: List[Issue] = []

        for output in total_output:
            issues.append(
                Issue(
                    output.source,
                    output.line,
                    self.get_name(),
                    output.type,
                    output.level,
                    output.message,
                    None,
                )
            )
        return issues



"""Apply markdownlint tool and gather results."""

import logging
import re
import subprocess
from typing import List, Match, Optional, Pattern

from statick_tool.issue import Issue
from statick_tool.package import Package
from statick_tool.tool_plugin import ToolPlugin


class MarkdownlintToolPlugin(ToolPlugin):  # type: ignore
    """Apply markdownlint tool and gather results."""

    def get_name(self) -> str:
        """Get name of tool."""
        return "markdownlint"

    # pylint: disable=too-many-locals
    def scan(self, package: Package, level: str) -> Optional[List[Issue]]:
        """Run tool and gather output."""
        tool_bin = "markdownlint"

        tool_config = ".markdownlintrc"
        user_config = self.plugin_context.config.get_tool_config(
            self.get_name(), level, "config"
        )
        if user_config is not None:
            tool_config = user_config

        format_file_name = self.plugin_context.resources.get_file(tool_config)
        flags: List[str] = []
        if format_file_name is not None:
            flags += ["-c", format_file_name]
        user_flags = self.get_user_flags(level)
        flags += user_flags

        files: List[str] = []
        if "md_src" in package:
            files += package["md_src"]

        total_output: List[str] = []

        for src in files:
            try:
                exe = [tool_bin] + flags + [src]
                output = subprocess.check_output(
                    exe, stderr=subprocess.STDOUT, universal_newlines=True
                )
                total_output.append(output)

            except subprocess.CalledProcessError as ex:
                if ex.returncode == 1:  # markdownlint returns 1 upon linting errors
                    total_output.append(ex.output)
                else:
                    logging.warning(
                        "%s failed! Returncode = %d", tool_bin, ex.returncode
                    )
                    logging.warning("%s exception: %s", self.get_name(), ex.output)
                    return None

            except OSError as ex:
                logging.warning("Couldn't find %s! (%s)", tool_bin, ex)
                return None

        for output in total_output:
            logging.debug("%s", output)

        with open(self.get_name() + ".log", "w", encoding="utf8") as fid:
            for output in total_output:
                fid.write(output)

        issues: List[Issue] = self.parse_output(total_output)
        return issues

    # pylint: enable=too-many-locals

    def parse_output(self, total_output: List[str]) -> List[Issue]:
        """Parse tool output and report issues."""
        markdownlint_re = r"(.+):(\d+)\s([^\s]+)\s(.+)"
        markdownlint_re_with_col = r"(.+):(\d+):(\d+)\s([^\s]+)\s(.+)"
        parse: Pattern[str] = re.compile(markdownlint_re)
        parse_with_col: Pattern[str] = re.compile(markdownlint_re_with_col)
        issues: List[Issue] = []

        for output in total_output:
            for line in output.split("\n"):
                match_with_col: Optional[Match[str]] = parse_with_col.match(line)
                if match_with_col:
                    issues.append(
                        Issue(
                            match_with_col.group(1),
                            match_with_col.group(2),
                            self.get_name(),
                            match_with_col.group(4),
                            3,
                            match_with_col.group(5),
                            None,
                        )
                    )
                else:
                    match: Optional[Match[str]] = parse.match(line)
                    if match:
                        issues.append(
                            Issue(
                                match.group(1),
                                match.group(2),
                                self.get_name(),
                                match.group(3),
                                3,
                                match.group(4),
                                None,
                            )
                        )
        return issues



"""Apply proselint tool and gather results.

Website: http://proselint.com/
Github: https://github.com/amperser/proselint

The tool uses the default proselint configuration file.
On Ubuntu this is at `~/.config/proselint/config`.

https://github.com/amperser/proselint#checks
"""
import json
import logging
from typing import Any, Dict, List, Optional

import proselint
from proselint.config import default as proselint_default

from statick_tool.issue import Issue
from statick_tool.package import Package
from statick_tool.tool_plugin import ToolPlugin


class ProselintToolPlugin(ToolPlugin):  # type: ignore
    """Apply proselint tool and gather results."""

    def get_name(self) -> str:
        """Get name of tool."""
        return "proselint"

    def scan(self, package: Package, level: str) -> Optional[List[Issue]]:
        """Run tool and gather output."""
        if "md_src" not in package or not package["md_src"]:
            return []

        files: List[str] = []
        if "md_src" in package:
            files += package["md_src"]

        # The JSON output does not include the filename so we have to run each file
        # one at a time, and store the output along with the filename in a dictionary.
        # The filename may be added to JSON output in the future:
        # https://github.com/amperser/proselint/issues/355
        output: Dict[str, Any] = {}
        for filename in files:
            with open(filename, encoding="utf8") as fid:
                errors = proselint.tools.errors_to_json(
                    proselint.tools.lint(fid, config=proselint_default)
                )
                output[filename] = errors

        for key, value in output.items():
            logging.debug("%s: %s", key, value)

        with open(self.get_name() + ".log", "w", encoding="utf8") as fid:
            for key, value in output.items():
                combined = key + value
                fid.write(combined)

        issues: List[Issue] = self.parse_output(output)
        return issues

    def parse_output(self, output: Dict[str, Any]) -> List[Issue]:
        """Parse tool output and report issues."""
        issues: List[Issue] = []
        for key, value in output.items():
            try:
                data = json.loads(value)["data"]["errors"]
            except KeyError as ex:
                logging.warning("%s exception: %s", self.get_name(), ex)
                continue
            for item in data:
                if (
                    "check" not in item
                    or "line" not in item
                    or "message" not in item
                    or "severity" not in item
                ):
                    logging.debug("  Found invalid proselint output: %s", item)
                    continue
                if item["severity"] == "suggestion":
                    warning_level = "1"
                elif item["severity"] == "warning":
                    warning_level = "3"
                elif item["severity"] == "error":
                    warning_level = "5"
                else:
                    warning_level = "3"

                issue = Issue(
                    key,
                    str(item["line"]),
                    self.get_name(),
                    item["check"],
                    warning_level,
                    item["message"],
                    None,
                )

                issues.append(issue)

        return issues



"""Setup."""


from setuptools import setup

with open("README.md", encoding="utf8") as fid:
    long_description = fid.read()  # pylint: disable=invalid-name

TEST_DEPS = [
    "mock",
    "pytest",
]

EXTRAS = {
    "test": TEST_DEPS,
}

setup(
    author="NIWC Pacific",
    name="statick-md",
    description="Statick analysis plugins for Markdown files.",
    version="0.1.0",
    packages=[
        "statick_tool",
        "statick_tool.plugins.discovery",
        "statick_tool.plugins.tool",
    ],
    package_dir={
        "statick_tool": ".",
        "statick_tool.plugins.discovery": "src/statick_md/plugins/discovery",
        "statick_tool.plugins.tool": "src/statick_md/plugins/tool",
    },
    package_data={
        "statick_tool": ["rsc/.*", "rsc/*"],
        "statick_tool.plugins.discovery": ["*.yapsy-plugin"],
        "statick_tool.plugins.tool": ["*.yapsy-plugin"],
    },
    long_description=long_description,
    long_description_content_type="text/markdown",
    install_requires=[
        "proselint",
        "restructuredtext-lint",
        "rstcheck",
        "sphinx",
        "statick",
    ],
    tests_require=TEST_DEPS,
    extras_require=EXTRAS,
    url="https://github.com/sscpac/statick-md",
    classifiers=[
        "License :: CC0 1.0 Universal (CC0 1.0) Public Domain Dedication",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Topic :: Software Development :: Testing",
    ],
)



"""Discover rst files to analyze."""

import logging
from collections import OrderedDict
from typing import List, Optional

from statick_tool.discovery_plugin import DiscoveryPlugin
from statick_tool.exceptions import Exceptions
from statick_tool.package import Package


class RstDiscoveryPlugin(DiscoveryPlugin):  # type: ignore
    """Discover rst files to analyze."""

    def get_name(self) -> str:
        """Get name of discovery type."""
        return "rst"

    def scan(
        self, package: Package, level: str, exceptions: Optional[Exceptions] = None
    ) -> None:
        """Scan package looking for rst files."""
        src_files: List[str] = []

        self.find_files(package)

        for file_dict in package.files.values():
            if file_dict["name"].endswith(".rst"):
                src_files.append(file_dict["path"])

        src_files = list(OrderedDict.fromkeys(src_files))

        logging.info("  %d rst files found.", len(src_files))
        if exceptions:
            original_file_count = len(src_files)
            src_files = exceptions.filter_file_exceptions_early(package, src_files)
            if original_file_count > len(src_files):
                logging.info(
                    "  After filtering, %d rst files will be scanned.",
                    len(src_files),
                )

        package["rst_src"] = src_files
